'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.runPending = exports.run = exports.getPending = undefined;

var _bluebird = require('bluebird');

var _bluebird2 = _interopRequireDefault(_bluebird);

var _changeCase = require('@rocketstation/change-case');

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _parseFunction = require('parse-function');

var _parseFunction2 = _interopRequireDefault(_parseFunction);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _pluralize = require('pluralize');

var _pluralize2 = _interopRequireDefault(_pluralize);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const getDirPath = (dir, model) => _path2.default.join(dir, model, 'migrations');
const getTableName = migrationTitle => (0, _pluralize2.default)((0, _changeCase.sl)(migrationTitle.split('.').slice(0, -1)));

const load = dir => {
  const modelsDir = _path2.default.join(dir, 'models');
  const result = {};
  _fs2.default.readdirSync(modelsDir).filter(item => _fs2.default.lstatSync(_path2.default.join(modelsDir, item)).isDirectory() && _fs2.default.existsSync(getDirPath(modelsDir, item))).forEach(item => {
    const migrationsDir = getDirPath(modelsDir, item);
    if (_fs2.default.existsSync(migrationsDir)) {
      _fs2.default.readdirSync(migrationsDir).filter(element => element.endsWith('.js')).forEach(element => {
        const [title] = element.split('.');
        const definition = require(_path2.default.join(migrationsDir, element));
        result[`${item}.${title}`] = [definition, ...(0, _parseFunction2.default)(definition).args];
      });
    }
  });
  return _bluebird2.default.resolve(result);
};

const getPending = async (connection, migrations) => {
  const data = await connection.query('select * from rstation_api_migrations');
  const executed = data.map(({ title }) => title);
  const createdTables = [];
  executed.forEach(item => {
    const table = getTableName(item);
    if (!createdTables.includes(table)) createdTables.push(table);
  });
  const pendingUnsorted = [];
  Object.keys(migrations).forEach(item => {
    if (!executed.includes(item)) {
      const table = getTableName(item);
      const method = migrations[item].up;
      const body = method.toString().replace(/\r?\n|\r/g, '').replace(/ /g, '');
      const references = [];
      const regex = /references:{model:'([a-z_]+)'|\.(addColumn|addIndex|removeColumn|changeColumn|renameColumn)\('([a-z_]+)'/g;
      let match = regex.exec(body);
      while (match) {
        let tableName;
        if (match[1]) {
          if (match[1] !== table) tableName = match[1];
        }
        if (match[3]) {
          tableName = match[3];
        }
        if (tableName && !references.includes(tableName)) references.push(tableName);
        match = regex.exec(body);
      }
      pendingUnsorted.push({ isCreateTable: body.includes('createTable'), method, table, references, title: item });
    }
  });
  let pending = pendingUnsorted.filter(({ references }) => references.length === 0);
  let i = 0;
  while (pending.length !== pendingUnsorted.length && i < 100) {
    pending.forEach(({ table }) => {
      if (!createdTables.includes(table)) createdTables.push(table);
    });
    pending = pending.concat(pendingUnsorted.filter(({ isCreateTable, table, references, title }) => {
      if (createdTables.includes(table) && !references.includes(table)) return false;
      if (references.includes(table) && pending.map(({ title }) => title).includes(title)) return false;
      if (references.includes(table) && !createdTables.includes(table) && !isCreateTable) return false;
      let flag = true;
      references.forEach(item => {
        flag = flag && (item === table || createdTables.includes(item));
      });
      return flag;
    }));
    i += 1;
  }
  if (i === 100) {
    console.log('Possible circular dependency found. Pending tables');
    console.log('SORTED:');
    console.log(pending.map(({ table, references }) => ({ table, references })));
    console.log('UNSORTED:');
    console.log(pendingUnsorted.map(({ table, references }) => ({ table, references })));
    return [];
  }
  return pending;
};

const run = async (connection, type, model, version = '', migrations, sequelize) => {
  await connection.query('create table if not exists "rstation_api_migrations" ("title" varchar(255) not null unique)');
  const title = `${model}.${version}`;
  const versions = Object.keys(migrations).filter(item => item.startsWith(title));
  if (versions.length === 0) return _bluebird2.default.reject(`no migrations found for '${model}' model`);
  if (version && !versions.includes(title)) return _bluebird2.default.reject(`no '${model}' model migration with '${version}' version found`);
  let toRun = [];
  let isToRun = () => false;
  if (type === 'up') isToRun = (list, item) => list.includes(item);
  if (type === 'down') isToRun = (list, item) => !list.includes(item);
  const pending = await getPending(connection, migrations);
  const pendingTitles = pending.map(({ title }) => title);
  if (version) {
    if (isToRun(pendingTitles, title)) toRun = [{ title, method: migrations[title][type] }];
  } else {
    toRun = versions.filter(item => isToRun(pendingTitles, item)).map(item => ({ title: item, method: migrations[item][type] }));
  }
  const queryInterface = sequelize.getQueryInterface();
  await _bluebird2.default.all(toRun.map(migration => migration.method(queryInterface)));
  if (type === 'up') await connection.query(`insert into rstation_api_migrations (title) values ${toRun.map(migration => `('${migration.title}')`).join(', ')};`);
  if (type === 'down') await connection.query(`delete from rstation_api_migrations where title in (${toRun.map(migration => `'${migration.title}'`).join(', ')});`);
};

const runPending = async (connection, migrations, sequelize) => {
  await connection.query('create table if not exists "rstation_api_migrations" ("title" varchar(255) not null unique)');
  const pending = await getPending(connection, migrations);
  const queryInterface = sequelize.getQueryInterface();
  if (pending.length > 0) {
    await _bluebird2.default.each(pending, item => item.method(queryInterface));
    await connection.query(`insert into rstation_api_migrations (title) values ${pending.map(migration => `('${migration.title}')`).join(', ')};`);
  }
};

exports.default = load;
exports.getPending = getPending;
exports.run = run;
exports.runPending = runPending;